/****************************************************************************/
/*  File:       MultipartResponseBody.java                                  */
/*  Author:     F. Georges - fgeorges.org                                   */
/*  Date:       2009-02-04                                                  */
/*  Tags:                                                                   */
/*      Copyright (c) 2009 Florent Georges (see end of file.)               */
/* ------------------------------------------------------------------------ */


package org.expath.httpclient.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.trans.XPathException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.james.mime4j.MimeException;
import org.apache.james.mime4j.field.AbstractField;
import org.apache.james.mime4j.parser.Field;
import org.apache.james.mime4j.parser.MimeTokenStream;
import org.expath.httpclient.ContentType;
import org.expath.httpclient.HeaderSet;
import org.expath.httpclient.HttpConnection;
import org.expath.httpclient.HttpResponse;
import org.expath.httpclient.HttpResponseBody;
import org.expath.httpclient.TreeBuilderHelper;

/**
 * TODO<doc>: ...
 *
 * @author Florent Georges
 * @date   2009-02-04
 */
public class MultipartResponseBody
        implements HttpResponseBody
{
    public MultipartResponseBody(InputStream in, ContentType type, HttpConnection conn, XPathContext ctxt)
            throws XPathException
    {
        myContentType = type;
        myParts = new ArrayList<HttpResponseBody>();
        Header h = conn.getResponseHeaders().getFirstHeader("Content-Type");
        if ( h == null ) {
            throw new XPathException("No content type");
        }
        myBoundary = type.getBoundary();
        if ( myBoundary == null ) {
            throw new XPathException("No boundary");
        }
        try {
            analyzeParts(in, h.getValue(), ctxt);
        }
        catch ( IOException ex ) {
            throw new XPathException("error reading the response stream", ex);
        }
    }

    public void setResponse(HttpResponse response)
    {
        // nothing
    }

    public void outputBody(TreeBuilderHelper b)
            throws XPathException
    {
        b.startElem("multipart");
        b.attribute("media-type", myContentType.getValue());
        b.attribute("boundary", myBoundary);
        b.startContent();
        for ( HttpResponseBody part : myParts ) {
            part.outputBody(b);
        }
        b.endElem();
    }

    public void addResultItems(List<Item> sequence)
            throws XPathException
    {
        for ( HttpResponseBody part : myParts ) {
            part.addResultItems(sequence);
        }
    }

//    private void analyzeParts_OLD_TOREMOVE(InputStream in, XPathContext ctxt)
//            throws IOException, XPathException
//    {
//        MultipartInputStream wrapper = new MultipartInputStream(in, myBoundary);
//        for ( int i = 1; wrapper.hasMorePart(); ++i ) {
//            HeaderSet headers = readHeaders(wrapper);
//            // Should not happen.
//            if ( headers.isEmpty() ) {
//                throw new XPathException("No header for part #" + i);
//            }
//            Header ct = headers.getFirstHeader("Content-Type");
//            ContentType part_type = new ContentType(ct);
//            myParts.add(BodyFactory.makeResponsePart(headers, wrapper, part_type, ctxt));
//            wrapper.nextPart();
//        }
//    }

//    private void analyzeParts(InputStream in, XPathContext ctxt)
//            throws IOException, XPathException
//    {
//        MultipartResponseHandler handler =
//                new MultipartResponseHandler(myParts, myContentType, myBoundary);
//        MimeStreamParser parser = new MimeStreamParser();
//        parser.setContentHandler(handler);
//        try {
//            parser.parse(in);
//        }
//        catch ( MimeException ex ) {
//            throw new XPathException("The response content is ill-formed.", ex);
//        }
//    }

    private void analyzeParts(InputStream in, String type, XPathContext ctxt)
            throws IOException, XPathException
    {
        MimeTokenStream parser = new MimeTokenStream();
        parser.parseHeadless(in, type);
        try {
            HeaderSet headers = null;
            for ( int state = parser.getState();
                  state != MimeTokenStream.T_END_OF_STREAM;
                  state = parser.next() )
            {
                if ( state == MimeTokenStream.T_START_HEADER ) {
                    headers = new HeaderSet();
                }
                handleParserState(parser, headers, ctxt);
            }
        }
        catch ( MimeException ex ) {
            throw new XPathException("The response content is ill-formed.", ex);
        }
    }

    private void handleParserState(MimeTokenStream parser, HeaderSet headers, XPathContext ctxt)
            throws XPathException
    {
        int state = parser.getState();
        if ( LOG.isDebugEnabled() ) {
            LOG.debug(MimeTokenStream.stateToString(state));
        }
        switch ( state ) {
            // It seems that in a headless parsing, END_HEADER appears
            // right after START_MESSAGE (without the corresponding
            // START_HEADER).  So if headers == null, we can just ignore
            // this state.
            case MimeTokenStream.T_END_HEADER:
                // TODO: Just ignore anyway...?
                break;
            case MimeTokenStream.T_FIELD:
                Field f = parser.getField();
                if ( LOG.isDebugEnabled() ) {
                    LOG.debug("  field: " + f);
                }
                headers.add(f.getName(), parseFieldBody(f));
                break;
            case MimeTokenStream.T_BODY:
                if ( LOG.isDebugEnabled() ) {
                    LOG.debug("  body desc: " + parser.getBodyDescriptor());
                }
                HttpResponseBody b = makeResponsePart(headers, parser, ctxt);
                myParts.add(b);
                break;
            // START_HEADER is handled in the calling analyzeParts()
            case MimeTokenStream.T_START_HEADER:
            case MimeTokenStream.T_END_BODYPART:
            case MimeTokenStream.T_END_MESSAGE:
            case MimeTokenStream.T_END_MULTIPART:
            case MimeTokenStream.T_EPILOGUE:
            case MimeTokenStream.T_PREAMBLE:
            case MimeTokenStream.T_START_BODYPART:
            case MimeTokenStream.T_START_MESSAGE:
            case MimeTokenStream.T_START_MULTIPART:
                // ignore
                break;
            // In a first time, take a very defensive approach, and
            // throw an error for all unexpected states, even if we
            // should discover slowly that we should probably just
            // ignore some of them.
            default:
                String s = MimeTokenStream.stateToString(state);
                throw new XPathException("Unknown parsing state: " + s);
        }
    }

    private String parseFieldBody(Field f)
            throws XPathException
    {
        try {
            String b = AbstractField.parse(f.getRaw()).getBody();
            if ( LOG.isDebugEnabled() ) {
                LOG.debug("Field: " + f.getName() + ": [" + b + "]");
            }
            return b;
        }
        catch ( MimeException ex ) {
            LOG.error("Field value parsing error (" + f + ")", ex);
            throw new XPathException("Field value parsing error (" + f + ")", ex);
        }
    }

    private HttpResponseBody makeResponsePart(HeaderSet headers, MimeTokenStream parser, XPathContext ctxt)
            throws XPathException
    {
        Header h = headers.getFirstHeader("Content-Type");
        if ( h == null ) {
            throw new XPathException("impossible to find the content type");
        }
        ContentType type = new ContentType(h);
        switch ( BodyFactory.parseType(type) ) {
            case XML: {
                // TODO: 'content_type' is the header Content-Type without any
                // param (i.e. "text/xml".)  Should we keep this, or put the
                // whole header (i.e. "text/xml; charset=utf-8")? (and for
                // other types as well...)
                Reader in = parser.getReader();
                return new XmlResponseBody(in, type, headers, ctxt, false);
            }
            case HTML: {
                Reader in = parser.getReader();
                return new XmlResponseBody(in, type, headers, ctxt, true);
            }
            case TEXT: {
                Reader in = parser.getReader();
                return new TextResponseBody(in, type, headers);
            }
            case BINARY: {
                InputStream in = parser.getInputStream();
                return new BinaryResponseBody(in, type, headers);
            }
            default:
                throw new XPathException("INTERNAL ERROR: cannot happen");
        }
    }

//    private HeaderSet readHeaders(MultipartInputStream in)
//            throws XPathException
//    {
//        try {
//            HeaderSet headers = new HeaderSet();
//            while ( true ) {
//                ByteArrayOutputStream buf = new ByteArrayOutputStream();
//                boolean seen_r = false;
//                int c = 0;
//                while ( (c = in.read()) != -1 ) {
//                    if ( seen_r && c == 10 ) { // 10 = \n
//                        int lwsp = in.readAhead();
//                        // A header value can be on several lines, if the first
//                        // char on the 2d line is either SPC or TAB.
//                        if ( lwsp == 32 || lwsp == 9 ) { // 32 = SPC, 9 = TAB
//                            in.read();
//                            continue;
//                        }
//                        else {
//                            break;
//                        }
//                    }
//                    if ( !seen_r && c == 13 ) { // 13 = \r
//                        seen_r = true;
//                    }
//                    else {
//                        buf.write(c);
//                    }
//                }
//                byte[] bytes = buf.toByteArray();
//                if ( bytes.length < 2 ) {
//                    break;
//                }
//                if ( bytes.length == 2 && bytes[0] == 13 && bytes[1] == 10 ) {
//                    break;
//                }
//                String line = new String(bytes, "US-ASCII").trim();
//                int colon = line.indexOf(':');
//                if ( colon < 1 ) {
//                    throw new XPathException("header wrong format: " + line);
//                }
//                String name = line.substring(0, colon);
//                String value = line.substring(colon + 1);
//                headers.add(name, value);
//            }
//            return headers;
//        }
//        catch ( UnsupportedEncodingException ex ) {
//            throw new XPathException("INTERNAL ERROR: encoding US-ASCII not supported");
//        }
//        catch ( IOException ex ) {
//            throw new XPathException("error reading multipart headers", ex);
//        }
//    }

    private List<HttpResponseBody> myParts;
    private ContentType myContentType;
    private String myBoundary;
    private static Log LOG = LogFactory.getLog(MultipartResponseBody.class);
}


/* ------------------------------------------------------------------------ */
/*  DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS COMMENT.               */
/*                                                                          */
/*  The contents of this file are subject to the Mozilla Public License     */
/*  Version 1.0 (the "License"); you may not use this file except in        */
/*  compliance with the License. You may obtain a copy of the License at    */
/*  http://www.mozilla.org/MPL/.                                            */
/*                                                                          */
/*  Software distributed under the License is distributed on an "AS IS"     */
/*  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See    */
/*  the License for the specific language governing rights and limitations  */
/*  under the License.                                                      */
/*                                                                          */
/*  The Original Code is: all this file.                                    */
/*                                                                          */
/*  The Initial Developer of the Original Code is Florent Georges.          */
/*                                                                          */
/*  Contributor(s): none.                                                   */
/* ------------------------------------------------------------------------ */
